# -*- coding: utf-8 -*-

import os
from cleo import InputOption, InputArgument, ListInput
from eloquent.migrations import MigrationCreator, DatabaseMigrationRepository
from .base_command import BaseCommand


class MigrateMakeCommand(BaseCommand):

    def configure(self):
        super(MigrateMakeCommand, self).configure()

        self.set_name('migrations:make')
        self.set_description('Create a new migration file')
        self.add_argument('name', InputArgument.REQUIRED, 'The name of the migration.')
        self.add_option('create', 'C', InputOption.VALUE_NONE,
                        'The table to be created.')
        self.add_option('table', 't', InputOption.VALUE_OPTIONAL,
                        'The table to migrate.')
        self.add_option('path', 'p', InputOption.VALUE_OPTIONAL,
                        'The path of migrations files.')

    def execute(self, i, o):
        """
        Executes the command.

        :type i: cleo.inputs.input.Input
        :type o: cleo.outputs.output.Output
        """
        super(MigrateMakeCommand, self).execute(i, o)

        creator = MigrationCreator()

        name = i.get_argument('name')
        table = i.get_option('table')
        create = bool(i.get_option('create'))

        if not table and create is not False:
            table = create

        path = i.get_option('path')
        if path is None:
            path = self._get_migration_path()

        file_ = self._write_migration(creator, name, table, create, path)

        o.writeln('<info>Create migration: <comment>%s</comment></info>' % file_)

    def _write_migration(self, creator, name, table, create, path):
        """
        Write the migration file to disk.
        """
        file_ = os.path.basename(creator.create(name, path, table, create))

        return file_



# -*- coding: utf-8 -*-

from cleo import InputOption
from eloquent.migrations import Migrator, DatabaseMigrationRepository
from .base_command import BaseCommand


class StatusCommand(BaseCommand):

    def configure(self):
        super(StatusCommand, self).configure()

        self.set_name('migrations:status')
        self.set_description('Show a list of migrations up/down')
        self.add_option('database', 'd', InputOption.VALUE_OPTIONAL,
                        'The database connection to use')
        self.add_option('path', 'p', InputOption.VALUE_OPTIONAL,
                        'The path of migrations files to be executed.')

    def execute(self, i, o):
        """
        Executes the command.

        :type i: cleo.inputs.input.Input
        :type o: cleo.outputs.output.Output
        """
        super(StatusCommand, self).execute(i, o)

        database = i.get_option('database')
        repository = DatabaseMigrationRepository(self._resolver, 'migrations')

        migrator = Migrator(repository, self._resolver)

        if not migrator.repository_exists():
            return o.writeln('<error>No migrations found</error>')

        self._prepare_database(migrator, database, i, o)

        path = i.get_option('path')

        if path is None:
            path = self._get_migration_path()

        ran = migrator.get_repository().get_ran()

        migrations = []
        for migration in migrator._get_migration_files(path):
            if migration in ran:
                migrations.append(['<fg=cyan>%s</>' % migration, '<info>Yes</info>'])
            else:
                migrations.append(['<fg=cyan>%s</>' % migration, '<fg=red>No</>'])

        if migrations:
            table = self.get_helper('table')
            table.set_headers(['Migration', 'Ran?'])
            table.set_rows(migrations)
            table.render(o)
        else:
            return o.writeln('<error>No migrations found</error>')

        for note in migrator.get_notes():
            o.writeln(note)

    def _prepare_database(self, migrator, database, i, o):
        migrator.set_connection(database)



# -*- coding: utf-8 -*-


class Platform(object):

    INTERNAL_TYPE_MAPPING = {}

    def get_default_value_declaration_sql(self, field):
        default = ''

        if not field.get('notnull'):
            default = ' DEFAULT NULL'

        if 'default' in field and field['default'] is not None:
            default = ' DEFAULT \'%s\'' % field['default']

            if 'type' in field:
                type = field['type']

                if type in ['integer', 'bigint', 'smallint']:
                    default = ' DEFAULT %s' % field['default']
                elif type in ['datetime', 'datetimetz'] \
                        and field['default'] in [self.get_current_timestamp_sql(), 'NOW', 'now']:
                    default = ' DEFAULT %s' % self.get_current_timestamp_sql()
                elif type in ['time'] \
                        and field['default'] in [self.get_current_time_sql(), 'NOW', 'now']:
                    default = ' DEFAULT %s' % self.get_current_time_sql()
                elif type in ['date'] \
                        and field['default'] in [self.get_current_date_sql(), 'NOW', 'now']:
                    default = ' DEFAULT %s' % self.get_current_date_sql()
                elif type in ['boolean']:
                    default = ' DEFAULT \'%s\'' % self.convert_booleans(field['default'])

        return default

    def convert_booleans(self, item):
        if isinstance(item, list):
            for i, value in enumerate(item):
                if isinstance(value, bool):
                    item[i] = int(value)
        elif isinstance(item, bool):
            item = int(item)

        return item

    def get_current_date_sql(self):
        return 'CURRENT_DATE'

    def get_current_time_sql(self):
        return 'CURRENT_TIME'

    def get_current_timestamp_sql(self):
        return 'CURRENT_TIMESTAMP'

    def get_sql_type_declaration(self, column):
        internal_type = column['type']

        return getattr(self, 'get_%s_type_sql_declaration' % internal_type)(column)

    def get_column_declaration_sql(self, name, field):
        if 'column_definition' in field:
            column_def = self.get_custom_type_declaration_sql(field)
        else:
            default = self.get_default_value_declaration_sql(field)

            charset = field.get('charset', '')
            if charset:
                charset = ' ' + self.get_column_charset_declaration_sql(charset)

            collation = field.get('collation', '')
            if charset:
                charset = ' ' + self.get_column_collation_declaration_sql(charset)

            notnull = field.get('notnull', '')
            if notnull:
                notnull = ' NOT NULL'
            else:
                notnull = ''

            unique = field.get('unique', '')
            if unique:
                unique = ' ' + self.get_unique_field_declaration_sql()
            else:
                unique = ''

            check = field.get('check', '')

            type_decl = self.get_sql_type_declaration(field)
            column_def = type_decl + charset + default + notnull + unique + check + collation

        return name + ' ' + column_def

    def get_custom_type_declaration_sql(self, column_def):
        return column_def['column_definition']

    def get_column_charset_declaration_sql(self, charset):
        return ''

    def get_column_collation_declaration_sql(self, collation):
        if self.supports_column_collation():
            return 'COLLATE %s' % collation

        return ''

    def supports_column_collation(self):
        return False

    def get_unique_field_declaration_sql(self):
        return 'UNIQUE'

    def get_string_type_sql_declaration(self, column):
        if 'length' not in column:
            column['length'] = self.get_varchar_default_length()

        fixed = column.get('fixed', False)

        if column['length'] > self.get_varchar_max_length():
            return self.get_text_type_sql_declaration(column)

        return self.get_varchar_type_declaration_sql_snippet(column['length'], fixed)

    def get_binary_type_sql_declaration(self, column):
        if 'length' not in column:
            column['length'] = self.get_binary_default_length()

        fixed = column.get('fixed', False)

        if column['length'] > self.get_binary_max_length():
            return self.get_blob_type_sql_declaration(column)

        return self.get_binary_type_declaration_sql_snippet(column['length'], fixed)

    def get_varchar_type_declaration_sql_snippet(self, length, fixed):
        raise NotImplementedError('VARCHARS not supported by Platform')

    def get_binary_type_declaration_sql_snippet(self, length, fixed):
        raise NotImplementedError('BINARY/VARBINARY not supported by Platform')

    def get_decimal_type_sql_declaration(self, column):
        if 'precision' not in column or not column['precision']:
            column['precision'] = 10

        if 'scale' not in column or not column['scale']:
            column['precision'] = 0

        return 'NUMERIC(%s, %s)' % (column['precision'], column['scale'])

    def get_varchar_default_length(self):
        return 255

    def get_varchar_max_length(self):
        return 4000

    def get_binary_default_length(self):
        return 255

    def get_binary_max_length(self):
        return 4000

    def get_column_options(self):
        return []

    def get_type_mapping(self, db_type):
        return self.INTERNAL_TYPE_MAPPING[db_type]



# -*- coding: utf-8 -*-

from .grammar import SchemaGrammar
from ..blueprint import Blueprint
from ...query.expression import QueryExpression
from ...support.fluent import Fluent


class SQLiteSchemaGrammar(SchemaGrammar):

    _modifiers = ['nullable', 'default', 'increment']

    _serials = ['big_integer', 'integer']

    def compile_rename_column(self, blueprint, command, connection):
        """
        Compile a rename column command.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :param command: The command
        :type command: Fluent

        :param connection: The connection
        :type connection: eloquent.connections.Connection

        :rtype: list
        """
        # The code is a little complex. It will propably change
        # if we support complete diffs in dbal
        sql = []

        schema = connection.get_schema_manager()
        table = self.get_table_prefix() + blueprint.get_table()

        column = connection.get_column(table, command.from_)

        columns = schema.list_table_columns(table)
        indexes = schema.list_table_indexes(table)
        foreign_keys = schema.list_table_foreign_keys(table)

        diff = self._get_renamed_diff(blueprint, command, column, schema)
        renamed_columns = diff.renamed_columns

        old_column_names = list(map(lambda x: x.get_name(), columns))

        # We build the new column names
        new_column_names = []
        for column_name in old_column_names:
            if column_name in renamed_columns:
                new_column_names.append(renamed_columns[column_name].get_name())
            else:
                new_column_names.append(column_name)

        # We create a temporary table and insert the data into it
        temp_table = '__temp__' + self.get_table_prefix() + blueprint.get_table()
        sql.append('CREATE TEMPORARY TABLE %s AS SELECT %s FROM %s'
                   % (temp_table, self.columnize(old_column_names), table))

        # We drop the current table
        sql += Blueprint(table).drop().to_sql(None, self)

        # Building the list a new columns
        new_columns = []
        for column in columns:
            for column_name, changed_column in renamed_columns.items():
                if column_name == column.get_name():
                    new_columns.append(changed_column)

        # Here we will try to rebuild a new blueprint to create a new table
        # with the original name
        new_blueprint = Blueprint(table)
        new_blueprint.create()
        primary = []
        for column in columns:
            # Mapping the database type to the blueprint type
            type = schema.get_database_platform().TYPE_MAPPING[column.get_type().lower()]

            # If the column is a primary, we will add it to the blueprint later
            if column.get_platform_option('pk'):
                primary.append(column.get_name())

            # If the column is not one that's been renamed we reinsert it into the blueprint
            if column.get_name() not in renamed_columns.keys():
                col = getattr(new_blueprint, type)(column.get_name())

                # If the column is nullable, we flag it
                if not column.get_notnull():
                    col.nullable()

                # If the column has a default value, we add it
                if column.get_default() is not None:
                    col.default(QueryExpression(column.get_default()))

        # Inserting the renamed columns into the blueprint
        for column in new_columns:
            type = schema.get_database_platform().TYPE_MAPPING[column.get_type().lower()]

            col = getattr(new_blueprint, type)(column.get_name())
            if not column.get_notnull():
                col.nullable()

            if column.get_default() is not None:
                col.default(QueryExpression(column.get_default()))

        # We add the primary keys
        if primary:
            new_blueprint.primary(primary)

        # We rebuild the indexes
        for index in indexes:
            index_columns = index['columns']
            new_index_columns = []
            index_name = index['name']

            for column_name in index_columns:
                if column_name in renamed_columns:
                    new_index_columns.append(renamed_columns[column_name].get_name())
                else:
                    new_index_columns.append(column_name)

            if index_columns != new_index_columns:
                index_name = None

            if index['unique']:
                new_blueprint.unique(new_index_columns, index_name)
            else:
                new_blueprint.index(index['columns'], index_name)

        for foreign_key in foreign_keys:
            fkey_from = foreign_key['from']
            if fkey_from in renamed_columns:
                fkey_from = renamed_columns[fkey_from].get_name()

            new_blueprint.foreign(fkey_from)\
                .references(foreign_key['to'])\
                .on(foreign_key['table'])\
                .on_delete(foreign_key['on_delete'])\
                .on_update(foreign_key['on_update'])

        # We create the table
        sql += new_blueprint.to_sql(None, self)

        # We reinsert the data into the new table
        sql.append('INSERT INTO %s (%s) SELECT %s FROM %s'
                   % (self.wrap_table(table),
                      ', '.join(new_column_names),
                      self.columnize(old_column_names),
                      self.wrap_table(temp_table)
                      ))

        # Finally we drop the temporary table
        sql += Blueprint(temp_table).drop().to_sql(None, self)

        return sql

    def compile_change(self, blueprint, command, connection):
        """
        Compile a change column command into a series of SQL statement.

        :param blueprint: The blueprint
        :type blueprint: eloquent.schema.Blueprint

        :param command: The command
        :type command: Fluent

        :param connection: The connection
        :type connection: eloquent.connections.Connection

        :rtype: list
        """
        sql = []

        schema = connection.get_schema_manager()
        table = self.get_table_prefix() + blueprint.get_table()

        columns = schema.list_table_columns(table)
        indexes = schema.list_table_indexes(table)
        foreign_keys = schema.list_table_foreign_keys(table)

        diff = self._get_changed_diff(blueprint, schema)
        blueprint_changed_columns = blueprint.get_changed_columns()
        changed_columns = diff.changed_columns

        temp_table = '__temp__' + self.get_table_prefix() + blueprint.get_table()
        sql.append('CREATE TEMPORARY TABLE %s AS SELECT %s FROM %s'
                   % (temp_table, self.columnize(list(map(lambda x: x.get_name(), columns))), table))
        sql += Blueprint(table).drop().to_sql(None, self)

        new_columns = []
        for column in columns:
            for column_name, changed_column in changed_columns.items():
                if column_name == column.get_name():
                    for blueprint_column in blueprint_changed_columns:
                        if blueprint_column.name == column_name:
                            new_columns.append(blueprint_column)
                            break

                    break

        new_blueprint = Blueprint(table)
        new_blueprint.create()
        primary = []
        new_column_names = []
        for column in columns:
            type = schema.get_database_platform().TYPE_MAPPING[column.get_type().lower()]

            if column.get_platform_option('pk'):
                primary.append(column.get_name())

            if column.get_name() not in changed_columns:
                col = getattr(new_blueprint, type)(column.get_name())
                if not column.get_notnull():
                    col.nullable()

                new_column_names.append(column.get_name())

        for column in new_columns:
            column.change = False
            new_blueprint._add_column(**column.get_attributes())
            new_column_names.append(column.name)

        if primary:
            new_blueprint.primary(primary)

        for index in indexes:
            if index['unique']:
                new_blueprint.unique(index['columns'], index['name'])
            else:
                new_blueprint.index(index['columns'], index['name'])

        for foreign_key in foreign_keys:
            new_blueprint.foreign(foreign_key['from'])\
                .references(foreign_key['to'])\
                .on(foreign_key['table'])\
                .on_delete(foreign_key['on_delete'])\
                .on_update(foreign_key['on_update'])

        sql += new_blueprint.to_sql(None, self)
        sql.append('INSERT INTO %s (%s) SELECT %s FROM %s'
                   % (self.wrap_table(table),
                      ', '.join(sorted(new_column_names)),
                      self.columnize(sorted(list(map(lambda x: x.get_name(), columns)))),
                      self.wrap_table(temp_table)
                      ))
        sql += Blueprint(temp_table).drop().to_sql(None, self)

        return sql

    def compile_table_exists(self):
        """
        Compile the query to determine if a table exists

        :rtype: str
        """
        return "SELECT * FROM sqlite_master WHERE type = 'table' AND name = ?"

    def compile_column_exists(self, table):
        """
        Compile the query to determine the list of columns.
        """
        return 'PRAGMA table_info(%s)' % table.replace('.', '__')

    def compile_create(self, blueprint, command, _):
        """
        Compile a create table command.
        """
        columns = ', '.join(self._get_columns(blueprint))

        sql = 'CREATE TABLE %s (%s' % (self.wrap_table(blueprint), columns)

        sql += self._add_foreign_keys(blueprint)

        sql += self._add_primary_keys(blueprint)

        return sql + ')'

    def _add_foreign_keys(self, blueprint):
        sql = ''

        foreigns = self._get_commands_by_name(blueprint, 'foreign')

        for foreign in foreigns:
            sql += self._get_foreign_key(foreign)

            if foreign.get('on_delete'):
                sql += ' ON DELETE %s' % foreign.on_delete

            if foreign.get('on_update'):
                sql += ' ON UPDATE %s' % foreign.on_delete

        return sql

    def _get_foreign_key(self, foreign):
        on = self.wrap_table(foreign.on)

        columns = self.columnize(foreign.columns)

        references = foreign.references
        if not isinstance(references, list):
            references = [references]

        on_columns = self.columnize(references)

        return ', FOREIGN KEY(%s) REFERENCES %s(%s)' % (columns, on, on_columns)

    def _add_primary_keys(self, blueprint):
        primary = self._get_command_by_name(blueprint, 'primary')

        if primary:
            columns = self.columnize(primary.columns)

            return ', PRIMARY KEY (%s)' % columns

        return ''

    def compile_add(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        columns = self.prefix_list('ADD COLUMN', self._get_columns(blueprint))

        statements = []

        for column in columns:
            statements.append('ALTER TABLE %s %s' % (table, column))

        return statements

    def compile_unique(self, blueprint, command, _):
        columns = self.columnize(command.columns)

        table = self.wrap_table(blueprint)

        return 'CREATE UNIQUE INDEX %s ON %s (%s)' % (command.index, table, columns)

    def compile_index(self, blueprint, command, _):
        columns = self.columnize(command.columns)

        table = self.wrap_table(blueprint)

        return 'CREATE INDEX %s ON %s (%s)' % (command.index, table, columns)

    def compile_foreign(self, blueprint, command, _):
        pass

    def compile_drop(self, blueprint, command, _):
        return 'DROP TABLE %s' % self.wrap_table(blueprint)

    def compile_drop_if_exists(self, blueprint, command, _):
        return 'DROP TABLE IF EXISTS %s' % self.wrap_table(blueprint)

    def compile_drop_column(self, blueprint, command, connection):
        # The code is a little complex. It will propably change
        # if we support complete diffs in dbal
        sql = []

        schema = connection.get_schema_manager()
        table = self.get_table_prefix() + blueprint.get_table()

        columns = schema.list_table_columns(table)
        indexes = schema.list_table_indexes(table)
        foreign_keys = schema.list_table_foreign_keys(table)

        diff = self._get_table_diff(blueprint, schema)

        for name in command.columns:
            column = connection.get_column(blueprint.get_table(), name)

            diff.removed_columns[name] = column

        removed_columns = diff.removed_columns

        old_column_names = list(map(lambda x: x.get_name(), columns))

        # We build the new column names
        new_column_names = []
        for column_name in old_column_names:
            if column_name not in removed_columns:
                new_column_names.append(column_name)

        # We create a temporary table and insert the data into it
        temp_table = '__temp__' + self.get_table_prefix() + blueprint.get_table()
        sql.append('CREATE TEMPORARY TABLE %s AS SELECT %s FROM %s'
                   % (temp_table, self.columnize(old_column_names), table))

        # We drop the current table
        sql += Blueprint(table).drop().to_sql(None, self)

        # Here we will try to rebuild a new blueprint to create a new table
        # with the original name
        new_blueprint = Blueprint(table)
        new_blueprint.create()
        primary = []
        for column in columns:
            # If the column is not one that's been removed we reinsert it into the blueprint
            if column.get_name() in new_column_names:
                # Mapping the database type to the blueprint type
                type = schema.get_database_platform().TYPE_MAPPING[column.get_type().lower()]

                # If the column is a primary, we will add it to the blueprint later
                if column.get_platform_option('pk'):
                    primary.append(column.get_name())

                col = getattr(new_blueprint, type)(column.get_name())

                # If the column is nullable, we flag it
                if not column.get_notnull():
                    col.nullable()

                # If the column has a default value, we add it
                if column.get_default() is not None:
                    col.default(QueryExpression(column.get_default()))

        # We add the primary keys
        if primary:
            new_blueprint.primary(primary)

        # We rebuild the indexes
        for index in indexes:
            index_columns = index['columns']
            new_index_columns = []
            index_name = index['name']

            removed = False
            for column_name in index_columns:
                if column_name not in removed_columns:
                    new_index_columns.append(column_name)
                else:
                    removed = True
                    break

            if removed:
                continue

            if index_columns != new_index_columns:
                index_name = None

            if index['unique']:
                new_blueprint.unique(new_index_columns, index_name)
            else:
                new_blueprint.index(index['columns'], index_name)

        for foreign_key in foreign_keys:
            fkey_from = foreign_key['from']
            if fkey_from in removed_columns:
                continue

            new_blueprint.foreign(fkey_from)\
                .references(foreign_key['to'])\
                .on(foreign_key['table'])\
                .on_delete(foreign_key['on_delete'])\
                .on_update(foreign_key['on_update'])

        # We create the table
        sql += new_blueprint.to_sql(None, self)

        # We reinsert the data into the new table
        sql.append('INSERT INTO %s (%s) SELECT %s FROM %s'
                   % (self.wrap_table(table),
                      self.columnize(new_column_names),
                      self.columnize(new_column_names),
                      self.wrap_table(temp_table)
                      ))

        # Finally we drop the temporary table
        sql += Blueprint(temp_table).drop().to_sql(None, self)

        return sql

    def compile_drop_unique(self, blueprint, command, _):
        return 'DROP INDEX %s' % command.index

    def compile_drop_index(self, blueprint, command, _):
        return 'DROP INDEX %s' % command.index

    def compile_rename(self, blueprint, command, _):
        from_ = self.wrap_table(blueprint)

        return 'ALTER TABLE %s RENAME TO %s' % (from_, self.wrap_table(command.to))

    def _type_char(self, column):
        return 'VARCHAR'

    def _type_string(self, column):
        return 'VARCHAR'

    def _type_text(self, column):
        return 'TEXT'

    def _type_medium_text(self, column):
        return 'TEXT'

    def _type_long_text(self, column):
        return 'TEXT'

    def _type_integer(self, column):
        return 'INTEGER'

    def _type_big_integer(self, column):
        return 'INTEGER'

    def _type_medium_integer(self, column):
        return 'INTEGER'

    def _type_tiny_integer(self, column):
        return 'TINYINT'

    def _type_small_integer(self, column):
        return 'INTEGER'

    def _type_float(self, column):
        return 'FLOAT'

    def _type_double(self, column):
        return 'FLOAT'

    def _type_decimal(self, column):
        return 'NUMERIC'

    def _type_boolean(self, column):
        return 'TINYINT'

    def _type_enum(self, column):
        return 'VARCHAR'

    def _type_json(self, column):
        return 'TEXT'

    def _type_date(self, column):
        return 'DATE'

    def _type_datetime(self, column):
        return 'DATETIME'

    def _type_time(self, column):
        return 'TIME'

    def _type_timestamp(self, column):
        return 'DATETIME'

    def _type_binary(self, column):
        return 'BLOB'

    def _modify_nullable(self, blueprint, column):
        if column.get('nullable'):
            return ' NULL'

        return ' NOT NULL'

    def _modify_default(self, blueprint, column):
        if column.get('default') is not None:
            return ' DEFAULT %s' % self._get_default_value(column.default)

        return ''

    def _modify_increment(self, blueprint, column):
        if column.type in self._serials and column.auto_increment:
            return ' PRIMARY KEY AUTOINCREMENT'

        return ''



# -*- coding: utf-8 -*-


class TableDiff(object):

    def __init__(self, table_name, added_columns=None,
                 changed_columns=None, removed_columns=None, added_indexes=None,
                 changed_indexes=None, removed_indexes=None, from_table=None):
        self.name = table_name
        self.new_name = False
        self.added_columns = added_columns or {}
        self.changed_columns = changed_columns or {}
        self.removed_columns = removed_columns or {}
        self.added_indexes = added_indexes or []
        self.changed_indexes = changed_indexes or []
        self.removed_indexes = removed_indexes or []
        self.added_foreign_keys = []
        self.changed_foreign_keys = []
        self.removed_foreign_keys = []
        self.renamed_columns = {}
        self.renamed_indexes = {}
        self.from_table = from_table



# -*- coding: utf-8 -*-

from .platform import Platform
from ..table import Table
from ..column import Column


class PostgresPlatform(Platform):

    INTERNAL_TYPE_MAPPING = {
        'smallint': 'smallint',
        'int2': 'smallint',
        'serial': 'integer',
        'serial4': 'integer',
        'int': 'integer',
        'int4': 'integer',
        'integer': 'integer',
        'bigserial': 'bigint',
        'serial8': 'bigint',
        'bigint': 'bigint',
        'int8': 'bigint',
        'bool': 'boolean',
        'boolean': 'boolean',
        'text': 'text',
        'tsvector': 'text',
        'varchar': 'string',
        'interval': 'string',
        '_varchar': 'string',
        'char': 'string',
        'bpchar': 'string',
        'inet': 'string',
        'date': 'date',
        'datetime': 'datetime',
        'timestamp': 'datetime',
        'timestamptz': 'datetimez',
        'time': 'time',
        'timetz': 'time',
        'float': 'float',
        'float4': 'float',
        'float8': 'float',
        'double': 'float',
        'double precision': 'float',
        'real': 'float',
        'decimal': 'decimal',
        'money': 'decimal',
        'numeric': 'decimal',
        'year': 'date',
        'uuid': 'guid',
        'bytea': 'blob'
    }

    def get_list_table_columns_sql(self, table):
        sql = """SELECT
                    a.attnum,
                    quote_ident(a.attname) AS field,
                    t.typname AS type,
                    format_type(a.atttypid, a.atttypmod) AS complete_type,
                    (SELECT t1.typname FROM pg_catalog.pg_type t1 WHERE t1.oid = t.typbasetype) AS domain_type,
                    (SELECT format_type(t2.typbasetype, t2.typtypmod) FROM
                      pg_catalog.pg_type t2 WHERE t2.typtype = 'd' AND t2.oid = a.atttypid) AS domain_complete_type,
                    a.attnotnull AS isnotnull,
                    (SELECT 't'
                     FROM pg_index
                     WHERE c.oid = pg_index.indrelid
                        AND pg_index.indkey[0] = a.attnum
                        AND pg_index.indisprimary = 't'
                    ) AS pri,
                    (SELECT pg_get_expr(adbin, adrelid)
                     FROM pg_attrdef
                     WHERE c.oid = pg_attrdef.adrelid
                        AND pg_attrdef.adnum=a.attnum
                    ) AS default,
                    (SELECT pg_description.description
                        FROM pg_description WHERE pg_description.objoid = c.oid AND a.attnum = pg_description.objsubid
                    ) AS comment
                    FROM pg_attribute a, pg_class c, pg_type t, pg_namespace n
                    WHERE %s
                        AND a.attnum > 0
                        AND a.attrelid = c.oid
                        AND a.atttypid = t.oid
                        AND n.oid = c.relnamespace
                    ORDER BY a.attnum""" % self.get_table_where_clause(table)

        return sql

    def get_list_table_indexes_sql(self, table):
        table = table.replace('.', '__')

        return 'PRAGMA index_list(\'%s\')' % table

    def get_list_table_foreign_keys_sql(self, table):
        return 'SELECT quote_ident(r.conname) AS name, ' \
               'pg_catalog.pg_get_constraintdef(r.oid, true) AS condef ' \
               'FROM pg_catalog.pg_constraint r ' \
               'WHERE r.conrelid = ' \
               '(' \
               'SELECT c.oid ' \
               'FROM pg_catalog.pg_class c, pg_catalog.pg_namespace n ' \
               'WHERE ' + self.get_table_where_clause(table) + ' AND n.oid = c.relnamespace' \
               ')' \
               ' AND r.contype = \'f\''

    def get_table_where_clause(self, table, class_alias='c', namespace_alias='n'):
        where_clause = namespace_alias + '.nspname NOT IN (\'pg_catalog\', \'information_schema\', \'pg_toast\') AND '
        if table.find('.') >= 0:
            split = table.split('.')
            schema, table = split[0], split[1]
            schema = "'%s'" % schema
        else:
            schema = 'ANY(string_to_array((select replace(replace(setting, \'"$user"\', user), \' \', \'\')' \
                     ' from pg_catalog.pg_settings where name = \'search_path\'),\',\'))'

        where_clause += '%s.relname = \'%s\' AND %s.nspname = %s' % (class_alias, table, namespace_alias, schema)

        return where_clause

    def get_alter_table_sql(self, diff):
        """
        Get the ALTER TABLE SQL statement

        :param diff: The table diff
        :type diff: eloquent.dbal.table_diff.TableDiff

        :rtype: list
        """
        sql = []

        for column_diff in diff.changed_columns.values():
            if self.is_unchanged_binary_column(column_diff):
                continue

            old_column_name = column_diff.old_column_name
            column = column_diff.column

            if any([column_diff.has_changed('type'),
                    column_diff.has_changed('precision'),
                    column_diff.has_changed('scale'),
                    column_diff.has_changed('fixed')]):
                query = 'ALTER ' + old_column_name + ' TYPE ' + self.get_sql_type_declaration(column.to_dict())
                sql.append('ALTER TABLE ' + diff.name + ' ' + query)

            if column_diff.has_changed('default') or column_diff.has_changed('type'):
                if column.get_default() is None:
                    default_clause = ' DROP DEFAULT'
                else:
                    default_clause = ' SET' + self.get_default_value_declaration_sql(column.to_dict())

                query = 'ALTER ' + old_column_name + default_clause
                sql.append('ALTER TABLE ' + diff.name + ' ' + query)

            if column_diff.has_changed('notnull'):
                op = 'DROP'
                if column.get_notnull():
                    op = 'SET'

                query = 'ALTER ' + old_column_name + ' ' + op + ' NOT NULL'
                sql.append('ALTER TABLE ' + diff.name + ' ' + query)

            if column_diff.has_changed('autoincrement'):
                if column.get_autoincrement():
                    seq_name = self.get_identity_sequence_name(diff.name, old_column_name)

                    sql.append('CREATE SEQUENCE ' + seq_name)
                    sql.append('SELECT setval(\'' + seq_name + '\', '
                               '(SELECT MAX(' + old_column_name + ') FROM ' + diff.name + '))')
                    query = 'ALTER ' + old_column_name + ' SET DEFAULT nextval(\'' + seq_name + '\')'
                    sql.append('ALTER TABLE ' + diff.name + ' ' + query)
                else:
                    query = 'ALTER ' + old_column_name + ' DROP DEFAULT'
                    sql.append('ALTER TABLE ' + diff.name + ' ' + query)

            if column_diff.has_changed('length'):
                query = 'ALTER ' + old_column_name + ' TYPE ' + self.get_sql_type_declaration(column.to_dict())
                sql.append('ALTER TABLE ' + diff.name + ' ' + query)

        for old_column_name, column in diff.renamed_columns.items():
            sql.append('ALTER TABLE ' + diff.name + ' '
                       'RENAME COLUMN ' + old_column_name + ' TO ' + column.get_name())

        return sql

    def is_unchanged_binary_column(self, column_diff):
        column_type = column_diff.column.get_type()

        if column_type not in ['blob', 'binary']:
            return False

        if isinstance(column_diff.from_column, Column):
            from_column = column_diff.from_column
        else:
            from_column = None

        if from_column:
            from_column_type = self.INTERNAL_TYPE_MAPPING[from_column.get_type()]

            if from_column_type in ['blob', 'binary']:
                return False

            return len([x for x in column_diff.changed_properties if x not in ['type', 'length', 'fixed']]) == 0

        if column_diff.has_changed('type'):
            return False

        return len([x for x in column_diff.changed_properties if x not in ['length', 'fixed']]) == 0

    def convert_booleans(self, item):
        if isinstance(item, list):
            for i, value in enumerate(item):
                if isinstance(value, bool):
                    item[i] = str(value).lower()
        elif isinstance(item, bool):
            item = str(item).lower()

        return item

    def get_boolean_type_sql_declaration(self, column):
        return 'BOOLEAN'

    def get_integer_type_sql_declaration(self, column):
        if column.get('autoincrement'):
            return 'SERIAL'

        return 'INT'

    def get_bigint_type_sql_declaration(self, column):
        if column.get('autoincrement'):
            return 'BIGSERIAL'

        return 'BIGINT'

    def get_smallint_type_sql_declaration(self, column):
        return 'SMALLINT'

    def get_guid_type_sql_declaration(self, column):
        return 'UUID'

    def get_datetime_type_sql_declaration(self, column):
        return 'TIMESTAMP(0) WITHOUT TIME ZONE'

    def get_datetimetz_type_sql_declaration(self, column):
        return 'TIMESTAMP(0) WITH TIME ZONE'

    def get_date_type_sql_declaration(self, column):
        return 'DATE'

    def get_time_type_sql_declaration(self, column):
        return 'TIME(0) WITHOUT TIME ZONE'

    def get_string_type_sql_declaration(self, column):
        length = column.get('length', '255')
        fixed = column.get('fixed')

        if fixed:
            return 'CHAR(%s)' % length
        else:
            return 'VARCHAR(%s)' % length

    def get_binary_type_sql_declaration(self, column):
        return 'BYTEA'

    def get_blob_type_sql_declaration(self, column):
        return 'BYTEA'

    def get_text_type_sql_declaration(self, column):
        return 'TEXT'

    def get_decimal_type_sql_declaration(self, column):
        if 'precision' not in column or not column['precision']:
            column['precision'] = 10

        if 'scale' not in column or not column['scale']:
            column['precision'] = 0

        return 'DECIMAL(%s, %s)' % (column['precision'], column['scale'])

    def get_float_type_sql_declaration(self, column):
        return 'DOUBLE PRECISION'

    def supports_foreign_key_constraints(self):
        return True



# -*- coding: utf-8 -*-

import simplejson as json


class Fluent(object):

    def __init__(self, **attributes):
        self.__attributes = {}

        for key, value in attributes.items():
            self.__attributes[key] = value

    def get(self, key, default=None):
        return self.__attributes.get(key, default)

    def get_attributes(self):
        return self.__attributes

    def to_dict(self):
        return self.__attributes

    def to_json(self):
        return json.dumps(self.to_dict())

    def __contains__(self, item):
        return item in self.__attributes

    def __getitem__(self, item):
        return self.__attributes[item]

    def __setitem__(self, key, value):
        self.__attributes[key] = value

    def __delitem__(self, key):
        del self.__attributes[key]

    def __dynamic(self, method):
        def call(*args, **kwargs):
            if len(args):
                self.__attributes[method] = args[0]
            else:
                self.__attributes[method] = True

            return self

        return call

    def __getattr__(self, item):
        if item in self.__attributes:
            return self.__attributes[item]

        return self.__dynamic(item)

    def __setattr__(self, key, value):
        if key.startswith(('_Fluent__', '_%s__' % self.__class__.__name__, '__')):
            super(Fluent, self).__setattr__(key, value)
        elif callable(getattr(self, key, None)):
            return super(Fluent, self).__setattr__(key, value)
        else:
            self.__attributes[key] = value

    def __delattr__(self, item):
        del self.__attributes[item]



# -*- coding: utf-8 -*-

import re
from .column import Column
from .schema_manager import SchemaManager
from .platforms.mysql_platform import MySqlPlatform


class MySqlSchemaManager(SchemaManager):

    def _get_portable_table_column_definition(self, table_column):
        db_type = table_column['type'].lower()
        match = re.match('(.+)\((.*)\).*', db_type)
        if match:
            db_type = match.group(1)

        if 'length' in table_column:
            length = table_column['length']
        else:
            if match and match.group(2) and ',' not in match.group(2):
                length = int(match.group(2))
            else:
                length = 0

        fixed = None

        if 'name' not in table_column:
            table_column['name'] = ''

        precision = None
        scale = None

        type = self._platform.get_type_mapping(db_type)

        if db_type in ['char', 'binary']:
            fixed = True
        elif db_type in ['float', 'double', 'real', 'decimal', 'numeric']:
            match = re.match('([A-Za-z]+\(([0-9]+),([0-9]+)\))', table_column['type'])
            if match:
                precision = match.group(1)
                scale = match.group(2)
                length = None
        elif db_type == 'tinytext':
            length = MySqlPlatform.LENGTH_LIMIT_TINYTEXT
        elif db_type == 'text':
            length = MySqlPlatform.LENGTH_LIMIT_TEXT
        elif db_type == 'mediumtext':
            length = MySqlPlatform.LENGTH_LIMIT_MEDIUMTEXT
        elif db_type == 'tinyblob':
            length = MySqlPlatform.LENGTH_LIMIT_TINYBLOB
        elif db_type == 'blob':
            length = MySqlPlatform.LENGTH_LIMIT_BLOB
        elif db_type == 'mediumblob':
            length = MySqlPlatform.LENGTH_LIMIT_MEDIUMBLOB
        elif db_type in ['tinyint', 'smallint', 'mediumint', 'int', 'bigint', 'year']:
            length = None

        if length is None or length == 0:
            length = None

        options = {
            'length': length,
            'unsigned': table_column['type'].find('unsigned') != -1,
            'fixed': fixed,
            'notnull': table_column['null'] != 'YES',
            'default': table_column.get('default'),
            'precision': None,
            'scale': None,
            'autoincrement': table_column['extra'].find('auto_increment') != -1
        }

        if scale is not None and precision is not None:
            options['scale'] = scale
            options['precision'] = precision

        column = Column(table_column['field'], type, options)

        if 'collation' in table_column:
            column.set_platform_option('collation', table_column['collation'])

        return column

    def get_database_platform(self):
        return MySqlPlatform()



# -*- coding: utf-8 -*-

from flexmock import flexmock, flexmock_teardown
from eloquent.connections import Connection
from eloquent.schema import SchemaBuilder
from .. import EloquentTestCase


class SchemaBuilderTestCase(EloquentTestCase):

    def tearDown(self):
        flexmock_teardown()

    def test_has_table_correctly_calls_grammar(self):
        connection = flexmock(Connection(None))
        grammar = flexmock()
        connection.should_receive('get_schema_grammar').and_return(grammar)
        builder = SchemaBuilder(connection)
        grammar.should_receive('compile_table_exists').once().and_return('sql')
        connection.should_receive('get_table_prefix').once().and_return('prefix_')
        connection.should_receive('select').once().with_args('sql', ['prefix_table']).and_return(['prefix_table'])

        self.assertTrue(builder.has_table('table'))



# -*- coding: utf-8 -*-


class ColumnDiff(object):

    def __init__(self, old_column_name, column, changed_properties=None, from_column=None):
        self.old_column_name = old_column_name
        self.column = column
        self.changed_properties = changed_properties
        self.from_column = from_column

    def has_changed(self, property_name):
        return property_name in self.changed_properties

    def get_old_column_name(self):
        return self.old_column_name



# -*- coding: utf-8 -*-

from .scope import Scope


class SoftDeletingScope(Scope):

    _extensions = ['force_delete', 'restore', 'with_trashed', 'only_trashed']

    def apply(self, builder, model):
        """
        Apply the scope to a given query builder.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder

        :param model: The model
        :type model: eloquent.orm.Model
        """
        builder.where_null(model.get_qualified_deleted_at_column())

        self.extend(builder)

    def remove(self, builder, model):
        """
        Remove the scope from a given query builder.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder

        :param model: The model
        :type model: eloquent.orm.Model
        """
        column = model.get_qualified_deleted_at_column()

        query = builder.get_query()

        wheres = []
        for where in query.wheres:
            # If the where clause is a soft delete date constraint,
            # we will remove it from the query and reset the keys
            # on the wheres. This allows the developer to include
            # deleted model in a relationship result set that is lazy loaded.
            if not self._is_soft_delete_constraint(where, column):
                wheres.append(where)

        query.wheres = wheres

    def extend(self, builder):
        """
        Extend the query builder with the needed functions.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        for extension in self._extensions:
            getattr(self, '_add_%s' % extension)(builder)

        builder.on_delete(self._on_delete)

    def _on_delete(self, builder):
        """
        The delete replacement function.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        column = self._get_deleted_at_column(builder)

        return builder.update({
            column: builder.get_model().fresh_timestamp()
        })

    def _get_deleted_at_column(self, builder):
        """
        Get the "deleted at" column for the builder.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder

        :rtype: str
        """
        if len(builder.get_query().joins) > 0:
            return builder.get_model().get_qualified_deleted_at_column()
        else:
            return builder.get_model().get_deleted_at_column()

    def _add_force_delete(self, builder):
        """
        Add the force delete extension to the builder.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        builder.macro('force_delete', self._force_delete)

    def _force_delete(self, builder):
        """
        The forece delete extension.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        return builder.get_query().delete()

    def _add_restore(self, builder):
        """
        Add the restore extension to the builder.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        builder.macro('restore', self._restore)

    def _restore(self, builder):
        """
        The restore extension.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        builder.with_trashed()

        return builder.update({
            builder.get_model().get_deleted_at_column(): None
        })

    def _add_with_trashed(self, builder):
        """
        Add the with-trashed extension to the builder.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        builder.macro('with_trashed', self._with_trashed)

    def _with_trashed(self, builder):
        """
        The with-trashed extension.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        self.remove(builder, builder.get_model())

        return builder

    def _add_only_trashed(self, builder):
        """
        Add the only-trashed extension to the builder.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        builder.macro('only_trashed', self._only_trashed)

    def _only_trashed(self, builder):
        """
        The only-trashed extension.

        :param builder: The query builder
        :type builder: eloquent.orm.builder.Builder
        """
        model = builder.get_model()

        self.remove(builder, model)

        builder.get_query().where_not_null(model.get_qualified_deleted_at_column())

    def _is_soft_delete_constraint(self, where, column):
        """
        Determine if the given where clause is a soft delete constraint.

        :param where: The where clause
        :type where: dict

        :param column: The column
        :type column: str

        :rtype: bool
        """
        return where['type'] == 'null' and where['column'] == column



# -*- coding: utf-8 -*-

from .platform import Platform
from ..table import Table
from ..column import Column


class MySqlPlatform(Platform):

    LENGTH_LIMIT_TINYTEXT = 255
    LENGTH_LIMIT_TEXT = 65535
    LENGTH_LIMIT_MEDIUMTEXT = 16777215

    LENGTH_LIMIT_TINYBLOB = 255
    LENGTH_LIMIT_BLOB = 65535
    LENGTH_LIMIT_MEDIUMBLOB = 16777215

    INTERNAL_TYPE_MAPPING = {
        'tinyint': 'boolean',
        'smallint': 'smallint',
        'mediumint': 'integer',
        'int': 'integer',
        'integer': 'integer',
        'bigint': 'bigint',
        'int8': 'bigint',
        'bool': 'boolean',
        'boolean': 'boolean',
        'tinytext': 'text',
        'mediumtext': 'text',
        'longtext': 'text',
        'text': 'text',
        'varchar': 'string',
        'string': 'string',
        'char': 'string',
        'date': 'date',
        'datetime': 'datetime',
        'timestamp': 'datetime',
        'time': 'time',
        'float': 'float',
        'double': 'float',
        'real': 'float',
        'decimal': 'decimal',
        'numeric': 'decimal',
        'year': 'date',
        'longblob': 'blob',
        'blob': 'blob',
        'mediumblob': 'blob',
        'tinyblob': 'blob',
        'binary': 'binary',
        'varbinary': 'binary',
        'set': 'simple_array'
    }

    def get_list_table_columns_sql(self, table, database=None):
        if database:
            database = "'%s'" % database
        else:
            database = 'DATABASE()'

        return 'SELECT COLUMN_NAME AS field, COLUMN_TYPE AS type, IS_NULLABLE AS `null`, ' \
               'COLUMN_KEY AS `key`, COLUMN_DEFAULT AS `default`, EXTRA AS extra, COLUMN_COMMENT AS comment, ' \
               'CHARACTER_SET_NAME AS character_set, COLLATION_NAME AS collation ' \
               'FROM information_schema.COLUMNS WHERE TABLE_SCHEMA = %s AND TABLE_NAME = \'%s\''\
               % (database, table)

    def get_list_table_indexes_sql(self, table, current_database=None):
        return 'SHOW INDEX FROM %s' % table

    def get_list_table_foreign_keys_sql(self, table, database=None):
        sql = ("SELECT DISTINCT k.`CONSTRAINT_NAME` AS `name`, k.`COLUMN_NAME`, k.`REFERENCED_TABLE_NAME`, "
               "k.`REFERENCED_COLUMN_NAME` /*!50116 , c.update_rule, c.delete_rule */ "
               "FROM information_schema.key_column_usage k /*!50116 "
               "INNER JOIN information_schema.referential_constraints c ON "
               "  c.constraint_name = k.constraint_name AND "
               "  c.table_name = '%s' */ WHERE k.table_name = '%s'" % (table, table))

        if database:
            sql += " AND k.table_schema = '%s' /*!50116 AND c.constraint_schema = '%s' */"\
                   % (database, database)

        sql += " AND k.`REFERENCED_COLUMN_NAME` IS NOT NULL"

        return sql

    def get_alter_table_sql(self, diff):
        """
        Get the ALTER TABLE SQL statement

        :param diff: The table diff
        :type diff: eloquent.dbal.table_diff.TableDiff

        :rtype: list
        """
        column_sql = []
        query_parts = []

        if diff.new_name is not False:
            query_parts.append('RENAME TO %s' % diff.new_name)

        # Added columns?

        # Removed columns?

        for column_diff in diff.changed_columns.values():
            column = column_diff.column
            column_dict = column.to_dict()

            # Don't propagate default value changes for unsupported column types.
            if column_diff.has_changed('default') \
                    and len(column_diff.changed_properties) == 1 \
                    and (column_dict['type'] == 'text' or column_dict['type'] == 'blob'):
                continue

            query_parts.append('CHANGE %s %s'
                               % (column_diff.get_old_column_name(),
                                  self.get_column_declaration_sql(column.get_name(), column_dict)))

        for old_column_name, column in diff.renamed_columns.items():
            column_dict = column.to_dict()
            query_parts.append('CHANGE %s %s'
                               % (self.quote(old_column_name),
                                  self.get_column_declaration_sql(self.quote(column.get_name()), column_dict)))

        sql = []

        if len(query_parts) > 0:
            sql.append('ALTER TABLE %s %s' % (diff.name, ', '.join(query_parts)))

        return sql

    def convert_booleans(self, item):
        if isinstance(item, list):
            for i, value in enumerate(item):
                if isinstance(value, bool):
                    item[i] = str(value).lower()
        elif isinstance(item, bool):
            item = str(item).lower()

        return item

    def get_boolean_type_sql_declaration(self, column):
        return 'TINYINT(1)'

    def get_integer_type_sql_declaration(self, column):
        return 'INT ' + self._get_common_integer_type_declaration_sql(column)

    def get_bigint_type_sql_declaration(self, column):
        return 'BIGINT ' + self._get_common_integer_type_declaration_sql(column)

    def get_smallint_type_sql_declaration(self, column):
        return 'SMALLINT ' + self._get_common_integer_type_declaration_sql(column)

    def get_guid_type_sql_declaration(self, column):
        return 'UUID'

    def get_datetime_type_sql_declaration(self, column):
        if 'version' in column and column['version'] == True:
            return 'TIMESTAMP'

        return 'DATETIME'

    def get_date_type_sql_declaration(self, column):
        return 'DATE'

    def get_time_type_sql_declaration(self, column):
        return 'TIME'

    def get_varchar_type_declaration_sql_snippet(self, length, fixed):
        if fixed:
            return 'CHAR(%s)' % length if length else 'CHAR(255)'
        else:
            return 'VARCHAR(%s)' % length if length else 'VARCHAR(255)'

    def get_binary_type_declaration_sql_snippet(self, length, fixed):
        if fixed:
            return 'BINARY(%s)' % (length or 255)
        else:
            return 'VARBINARY(%s)' % (length or 255)

    def get_text_type_sql_declaration(self, column):
        length = column.get('length')
        if length:
            if length <= self.LENGTH_LIMIT_TINYTEXT:
                return 'TINYTEXT'

            if length <= self.LENGTH_LIMIT_TEXT:
                return 'TEXT'

            if length <= self.LENGTH_LIMIT_MEDIUMTEXT:
                return 'MEDIUMTEXT'

        return 'LONGTEXT'

    def get_blob_type_sql_declaration(self, column):
        length = column.get('length')
        if length:
            if length <= self.LENGTH_LIMIT_TINYBLOB:
                return 'TINYBLOB'

            if length <= self.LENGTH_LIMIT_BLOB:
                return 'BLOB'

            if length <= self.LENGTH_LIMIT_MEDIUMBLOB:
                return 'MEDIUMBLOB'

        return 'LONGBLOB'

    def get_decimal_type_sql_declaration(self, column):
        decl = super(MySqlPlatform, self).get_decimal_type_sql_declaration(column)

        return decl + self.get_unsigned_declaration(column)

    def get_unsigned_declaration(self, column):
        if column.get('unsigned'):
            return ' UNSIGNED'

        return ''

    def _get_common_integer_type_declaration_sql(self, column):
        autoinc = ''
        if column.get('autoincrement'):
            autoinc = ' AUTO_INCREMENT'

        return self.get_unsigned_declaration(column) + autoinc

    def get_float_type_sql_declaration(self, column):
        return 'DOUBLE PRECISION' + self.get_unsigned_declaration(column)

    def supports_foreign_key_constraints(self):
        return True

    def supports_column_collation(self):
        return False

    def quote(self, name):
        return '`%s`' % name.replace('`', '``')



# -*- coding: utf-8 -*-

import os
from cleo import Command, InputOption, ListInput
from eloquent import DatabaseManager


class BaseCommand(Command):

    def __init__(self):
        super(BaseCommand, self).__init__()

        self._resolver = {}

    def configure(self):
        self.add_option('config', 'c', InputOption.VALUE_REQUIRED, 'The config file path')

    def execute(self, i, o):
        """
        Executes the command.

        :type i: cleo.inputs.input.Input
        :type o: cleo.outputs.output.Output
        """
        config = self._get_config(i)
        self._resolver = DatabaseManager(config)

    def call(self, name, options=None, o=None):
        """
        Call another command.

        :param name: The command name
        :type name: str

        :param options: The options
        :type options: list or None

        :param o: The output
        :type o: cleo.outputs.output.Output
        """
        if options is None:
            options = []

        command = self.get_application().find(name)

        options = [('command', command.get_name())] + options

        return command.run(ListInput(options), o)

    def _get_migration_path(self):
        return os.path.join(os.getcwd(), 'migrations')

    def _get_config(self, i):
        """
        Get the config.

        :type i: cleo.inputs.input.Input

        :rtype: dict
        """
        variables = {}
        if not i.get_option('config'):
            raise Exception('The --config|-c option is missing.')

        with open(i.get_option('config')) as fh:
            exec(fh.read(), {}, variables)

        return variables['DATABASES']



# -*- coding: utf-8 -*-

import re
from .table import Table
from .column import Column
from .schema_manager import SchemaManager
from .platforms.postgres_platform import PostgresPlatform


class PostgresSchemaManager(SchemaManager):

    def _get_portable_table_column_definition(self, table_column):
        if table_column['type'].lower() == 'varchar' or table_column['type'] == 'bpchar':
            length = re.sub('.*\(([0-9]*)\).*', '\\1', table_column['complete_type'])
            table_column['length'] = length

        autoincrement = False
        match = re.match("^nextval\('?(.*)'?(::.*)?\)$", str(table_column['default']))
        if match:
            table_column['sequence'] = match.group(1)
            table_column['default'] = None
            autoincrement = True

        match = re.match("^'?([^']*)'?::.*$", str(table_column['default']))
        if match:
            table_column['default'] = match.group(1)

        if str(table_column['default']).find('NULL') == 0:
            table_column['default'] = None

        if 'length' in table_column:
            length = table_column['length']
        else:
            length = None

        if length == '-1' and 'atttypmod' in table_column:
            length = table_column['atttypmod'] - 4

        if length is None or int(length) <= 0:
            length = None

        fixed = None

        if 'name' not in table_column:
            table_column['name'] = ''

        precision = None
        scale = None

        db_type = table_column['type'].lower()

        type = self._platform.get_type_mapping(db_type)

        if db_type in ['smallint', 'int2']:
            length = None
        elif db_type in ['int', 'int4', 'integer']:
            length = None
        elif db_type in ['int8', 'bigint']:
            length = None
        elif db_type in ['bool', 'boolean']:
            if table_column['default'] == 'true':
                table_column['default'] = True

            if table_column['default'] == 'false':
                table_column['default'] = False

            length = None
        elif db_type == 'text':
            fixed = False
        elif db_type in ['varchar', 'interval', '_varchar']:
            fixed = False
        elif db_type in ['char', 'bpchar']:
            fixed = True
        elif db_type in ['float', 'float4', 'float8',
                         'double', 'double precision',
                         'real', 'decimal', 'money', 'numeric']:
            match = re.match('([A-Za-z]+\(([0-9]+),([0-9]+)\))', table_column['complete_type'])
            if match:
                precision = match.group(1)
                scale = match.group(2)
                length = None
        elif db_type == 'year':
            length = None

        if table_column['default']:
            match = re.match("('?([^']+)'?::)", str(table_column['default']))
            if match:
                table_column['default'] = match.group(1)

        options = {
            'length': length,
            'notnull': table_column['isnotnull'],
            'default': table_column['default'],
            'primary': table_column['pri'] == 't',
            'precision': precision,
            'scale': scale,
            'fixed': fixed,
            'unsigned': False,
            'autoincrement': autoincrement
        }

        column = Column(table_column['field'], type, options)

        return column

    def get_database_platform(self):
        return PostgresPlatform()
